const NR_STEPS = 100;
const MAX_TRANSLATION_X = 2;
const MAX_TRANSLATION_Y = 2;
const TRANSLATION_RANGE = [[-2,2],[-2,2]];
const MIN_TRANSLATION_X = -2;
const MIN_TRANSLATION_Y = -2;
const MIN_ROTATION = 0.001;
const MAX_ROTATION = 0.01;
const ROTATION_RANGE = [0.001, 0.01];
//const FRAG_SIZE = new FragmentSize([[-6,6],[-6,6]]);
const MAX_FRAG_SIZE = new FragmentSize([[SLIDER_MIN, SLIDER_MAX],[SLIDER_MIN, SLIDER_MAX]]);
//const RESHAPE_RANGE = [-3,3];
const RESHAPE_RANGE = [-10,10];
const RELAX_RANGE = [2,3];
const PERTURB_RANGE = [-20,20];
const PERTURB_FRAC_RANGE = [-0.02,0.02];

var CURR_STEP = 0;

function transform(transfCanvas, transfGraph, settings) {
  //right_close();
  var checkboxes = document.getElementsByName("transformations"); 
  if (checkboxes[0].checked) {
    transfGraph.resetFInfo();
    runTranslate(transfCanvas, transfGraph);    
  }
  else if (checkboxes[1].checked) {
    runRotate(transfCanvas, transfGraph);
  }
  else if (checkboxes[2].checked) {
    transfGraph.resetFInfo();
    reshape(transfCanvas, transfGraph);
  }
  else if (checkboxes[3].checked) {
    transfGraph.resetFInfo();
    relax(transfCanvas, transfGraph);
  }
  else if (checkboxes[4].checked) {
    transfGraph.resetFInfo();
    perturbCell(transfCanvas, transfGraph);
  }
  else if (checkboxes[5].checked) {
    transfGraph.resetFInfo();
    perturbVOrbits(transfCanvas, transfGraph);
  }
}
/*==========================================================================*/
/*============================== TRANSLATION ===============================*/
/*==========================================================================*/
//https://www.w3schools.com/jsref/met_win_setinterval.asps
function runTranslate(transfCanvas, transfGraph) {
  if (transfGraph.vertices.length < 1) {
    alert("The transformation graph has no vertices. A translation of the vertices cannot be performed");
    return;
  }
  document.getElementById("transf_button").disabled = true;
  //making the fragment
  setNewRightFragmentSize(MAX_FRAG_SIZE);
  transfGraph.makeNewFragment(right_fragSize);
  transfGraph.computeLattice(right_fragSize);
  //computing a random translation to be applied at every step
  var translation = new Coordinate(getRandom(TRANSLATION_RANGE[0]), getRandom(TRANSLATION_RANGE[1]));
  //translation_step recursively translates the transfGraph NR_STEPS times
  //requestAnimationFrame makes the animation look smooth
  requestAnimationFrame(function(timestamp) {
    translation_step(transfCanvas, transfGraph, translation);
  });
}

function translation_step(transfCanvas, transfGraph, translation) {
  if (CURR_STEP < NR_STEPS) {
    translatePoints(transfGraph.vertices, translation);
    transfGraph.computeSpan();
    drawGraph(transfCanvas, transfGraph, settings);
    //Recursion
    requestAnimationFrame(function(timestamp) {
        CURR_STEP++;
        translation_step(transfCanvas, transfGraph, translation);
    });
  }
  //Base Case
  else {
    CURR_STEP = 0;
    transfGraph.recomputeVertexReps();//recompute the fractional vertex coordinates of the relaxed cell
    transfGraph.recomputeEdgeRepShifts();//recompute the shift of the edge representatives
    transfGraph.makeNewFragment(right_fragSize); //recompute the fragment with the updated vertex IDs and shifts
    transfGraph.computeSpan();
    document.getElementById("transf_button").disabled = false;
    drawGraph(transfCanvas, transfGraph, settings);
    printGraph(transfGraph);
  }
}
/*===========================================================================*/
/*================================ ROTATION =================================*/
/*===========================================================================*/
//Rotate axes and graph along the origin
function runRotate(transfCanvas, transfGraph) {
  if (transfGraph.axes.length < 2) {
    alert("A rotation cannot be performed on a graph without axes.");
    return;
  }
  document.getElementById("transf_button").disabled = true;
  var rotation_radians = getRandom(ROTATION_RANGE);
  //rotation_step recursively rotates the graph by rotation_radians NR_STEPS times
  requestAnimationFrame(function(timestamp) {
    rotation_step(transfCanvas, transfGraph, rotation_radians);
  });
}
function rotation_step(transfCanvas, transfGraph, rotation) {
  if (CURR_STEP < NR_STEPS) {
    for (let i = 0; i < transfGraph.axes.length; i++) {
      //using (0,0) for the center of rotation because the axes are vectors
      transfGraph.axes[i] = rotatePt((new Coordinate(0,0)), transfGraph.axes[i], rotation);
      transfGraph.computeLattice(right_fragSize);
      transfGraph.makeNewFragment(right_fragSize, false);//update vertex cartesian coordinates
    }
    transfGraph.computeSpan();
    drawGraph(transfCanvas, transfGraph, settings);
    //Recursion
    requestAnimationFrame(function(timestamp) {
        CURR_STEP++;
        rotation_step(transfCanvas, transfGraph, rotation);
    });
  }
  //Base Case
  else {
    CURR_STEP = 0;
    document.getElementById("transf_button").disabled = false;
  }
}
/*==========================================================================*/
/*================================ RESHAPE =================================*/
/*==========================================================================*/
function reshape(transfCanvas, transfGraph) {
  if (transfGraph.axes.length < 2) {
    alert("Axes must be chosen before reshaping the cell.");
    return;
  }
  //making the fragment
  setNewRightFragmentSize(MAX_FRAG_SIZE);

  var new_xaxis_frac = [];
  var new_yaxis_frac = [];
  var new_xaxis = [];
  var new_yaxis = [];
  var new_yaxis_sols = [];
  var good_new_yaxis_sols = [];
  var count1 = 0;
  while (count1 < 1000 && good_new_yaxis_sols.length === 0) { 
    //if no valid y axis found fond for the chosen x axis, try again
    var count2 = 0;
    new_yaxis_sols = [];
    while (new_yaxis_sols.length === 0 && count2 < 1000) {
      new_xaxis_frac = new Coordinate(getRandomInt(RESHAPE_RANGE),getRandomInt(RESHAPE_RANGE));
      new_yaxis_sols = getMatchingY(new_xaxis_frac);
      count2++;
    }
    if (count2 === 1001) {
      alert("Reshape failed unexpectedly. Click 'Transform' to try again.");
      return;
    }
    good_new_yaxis_sols = [];
    new_xaxis = fracToCartCoord(transfGraph, new_xaxis_frac);
    new_xaxis.subtract(transfGraph.origin);//to get the vector
    //check if solutions found have a bad angle.
    for (let i = 0; i < new_yaxis_sols.length; i ++) {
      new_yaxis_frac = new_yaxis_sols[i];
      new_yaxis = fracToCartCoord(transfGraph, new_yaxis_frac);
      new_yaxis.subtract(transfGraph.origin);
      var angle = Math.abs(getAngle2DDegrees(new_xaxis, new_yaxis));
      //If the angle is near 0 or 180 the cell deforms into a line
      if ((angle > 30 && angle < 150) || (angle > 210 && angle < 330))
        good_new_yaxis_sols.push(new_yaxis);
    }
    count1++;
  }
  if (good_new_yaxis_sols === 0) {
    alert("Reshape failed unexpectedly. Click 'Transform' to try again.");
    return;
  }
  
  //https://stackoverflow.com/questions/5915096/get-a-random-item-from-a-javascript-array
  new_yaxis = good_new_yaxis_sols[Math.floor(Math.random()*good_new_yaxis_sols.length)];
  transfGraph.axes = [new_xaxis,new_yaxis];
  
  transfGraph.computeLattice(right_fragSize);
  transfGraph.recomputeVertexReps();//recompute the fractional vertex coordinates
  transfGraph.recomputeEdgeRepShifts(right_fragSize);//recompute the shift of the edge representatives
  transfGraph.makeNewFragment(right_fragSize); //recompute the fragment with the updated vertex IDs and shifts
  transfGraph.computeSpan();
  drawGraph(transfCanvas, transfGraph, settings);
}
/*========================================================================*/
/*================================ RELAX =================================*/
/*========================================================================*/
function relax(transfCanvas, transfGraph) {
  if (transfGraph.axes.length < 2) {
    alert("Axes must be chosen before relaxing the cell.");
    return;
  }
  //making the fragment
  setNewRightFragmentSize(MAX_FRAG_SIZE);

  var new_xaxis_frac = [];
  var new_yaxis_frac = [];
  var new_xaxis = [];
  var new_yaxis = [];
  var new_yaxis_sols = [];
  var good_new_yaxis_sols = [];
  var count1 = 0;
  var relaxation_factor;
  while (count1 < 1000 && good_new_yaxis_sols.length === 0) { 
    //if no valid y axis found fond for the chosen x axis, try again
    var count2 = 0;
    new_yaxis_sols = [];
    while (new_yaxis_sols.length === 0 && count2 < 1000) {
      relaxation_factor = getRandomInt(RELAX_RANGE);
      new_xaxis_frac = new Coordinate(getRandomInt(RESHAPE_RANGE),getRandomInt(RESHAPE_RANGE));
      new_yaxis_sols = getMatchingYRelax(new_xaxis_frac, relaxation_factor);
      count2++;
    }
    if (count2 === 1001) {
      alert("Relax failed unexpectedly. Click 'Transform' to try again.");
      return;
    }
    good_new_yaxis_sols = [];
    new_xaxis = fracToCartCoord(transfGraph, new_xaxis_frac);
    new_xaxis.subtract(transfGraph.origin);//to get the vector
    //check if solutions found have a bad angle.
    for (let i = 0; i < new_yaxis_sols.length; i ++) {
      new_yaxis_frac = new_yaxis_sols[i];
      new_yaxis = fracToCartCoord(transfGraph, new_yaxis_frac);
      new_yaxis.subtract(transfGraph.origin);
      var angle = Math.abs(getAngle2DDegrees(new_xaxis, new_yaxis));
      //If the angle is near 0 or 180 the cell deforms into a line
      if ((angle > 30 && angle < 150) || (angle > 210 && angle < 330))
        good_new_yaxis_sols.push(new_yaxis);
    }
    count1++;
  }
  if (good_new_yaxis_sols === 0) {
    alert("Relax failed unexpectedly. Click 'Transform' to try again.");
    return;
  }
  new_yaxis = good_new_yaxis_sols[Math.floor(Math.random()*good_new_yaxis_sols.length)];
  transfGraph.axes = [new_xaxis,new_yaxis];
  //transfGraph.axes[1].multiply(2);
  
  transfGraph.computeLattice(right_fragSize);
  transfGraph.recomputeRelaxedVertexReps();//recompute the fractional vertex coordinates
  transfGraph.recomputeRelaxedEdgeReps();//recompute the shift of the edge representatives
  transfGraph.makeNewFragment(right_fragSize); //recompute the fragment with the updated vertex IDs and shifts
  transfGraph.computeSpan();
  drawGraph(transfCanvas, transfGraph, settings);
}
/*==========================================================================*/
/*================================ PERTURBATIONS =================================*/
/*==========================================================================*/
function perturbCell(transfCanvas, transfGraph) {
  if (transfGraph.axes.length < 2) {
    alert("Axes must be chosen before perturbing the cell.");
    return;
  }
  
  transfGraph.perturbAxes();
  
  transfGraph.computeLattice(right_fragSize);
  transfGraph.makeNewFragment(right_fragSize, false);
  transfGraph.computeSpan();
  drawGraph(transfCanvas, transfGraph, settings);
}
function perturbVOrbits(transfCanvas, transfGraph) {
  if (transfGraph.vorbits.length < 1) {
    alert("The transformation graph has no vertices. A perturbation of vertex orbits cannot be performed");
    return;
  }
  transfGraph.perturbVOrbits();
  
  transfGraph.computeLattice(right_fragSize);
  transfGraph.makeNewFragment(right_fragSize, false);
  transfGraph.computeSpan();
  drawGraph(transfCanvas, transfGraph, settings);
}
/*==========================================================================*/
/*================================ HELPERS =================================*/
/*==========================================================================*/
function getMatchingY(vector) {
  var solutions = [];
  for (let i = RESHAPE_RANGE[0]; i < RESHAPE_RANGE[1]; i++) {
    if(((1 - i * vector.x) % vector.y) === 0)
      solutions.push(new Coordinate((1 - i * vector.x) / -vector.y, i));
  }
  return solutions;
}
function getMatchingYRelax(vector, size) {
  var solutions = [];
  for (let i = RESHAPE_RANGE[0]; i < RESHAPE_RANGE[1]; i++) {
    if(((size - i * vector.x) % vector.y) === 0)
      solutions.push(new Coordinate((size - i * vector.x) / -vector.y, i));
  }
  return solutions;
}